#include <engine/3d/render.h>

#include <engine/3d/bezier.h>
#include <engine/allocmem.h>
#include <engine/asset.h>
#include <engine/display.h>
#include <engine/math.h>

#include <INLINE_C.H>

//#define WOBBLE_SUBDIV /* <----- this runs like shit, but it will get rid of seams between edges of polygons of differing subdivisions */

#define MAX_DIVBUF_F3		1024
#define MAX_DIVBUF_F4		7000
#define MAX_DIVBUF_FT3		1024
#define MAX_DIVBUF_FT4		2048
#define MAX_DIVBUF_G3		128
#define MAX_DIVBUF_G4		128
#define MAX_DIVBUF_GT3		128
#define MAX_DIVBUF_GT4		128

static long saveotz[MAX_OTZBUF];
static long saveotz_type[MAX_OTZBUF];
static long saveotz_is_visble[MAX_OTZBUF];

struct
{
	/* index 0 of all prim buffers is used as a guide for subdivision in mtd_add */
	POLY_F3		f3[MAX_DIVBUF_F3];
	POLY_F4		f4[MAX_DIVBUF_F4];
	POLY_FT3	ft3[MAX_DIVBUF_FT3];
	POLY_FT4	ft4[MAX_DIVBUF_FT4];
	POLY_G3		g3[MAX_DIVBUF_G3];
	POLY_G4		g4[MAX_DIVBUF_G4];
	POLY_GT3	gt3[MAX_DIVBUF_GT3];
	POLY_GT4	gt4[MAX_DIVBUF_GT4];

	DIVPOLYGON3	div3;
	DIVPOLYGON4	div4;

	char		*ptr[MAX_UNIQUE_POLY];
	unsigned int	idx[MAX_UNIQUE_POLY];
}divbuf;

extern unsigned long *RCpolyF3();
extern unsigned long *RCpolyF4();
extern unsigned long *RCpolyFT3();
extern unsigned long *RCpolyFT4();
extern unsigned long *RCpolyG3();
extern unsigned long *RCpolyG4();
extern unsigned long *RCpolyGT3();
extern unsigned long *RCpolyGT4();

static MATRIX *(*rotmat[3])(SVECTOR *r, MATRIX *m) = {
	RotMatrix_gte,
	RotMatrixYXZ_gte,
	RotMatrixZYX_gte
};

static MATRIX *(*rotmat_precise[3])(SVECTOR *r, MATRIX *m) = {
	RotMatrix,
	RotMatrixYXZ,
	RotMatrixZYX
};

static int (*rottbl[2])() = {
	rotavg3,
	rotavg4
};

static unsigned long *(*rcptbl[MAX_UNIQUE_POLY])() = {
	RCpolyF3, RCpolyF4, RCpolyFT3, RCpolyFT4,
	RCpolyG3, RCpolyG4, RCpolyGT3, RCpolyGT4
};

static unsigned int apb_size[MAX_UNIQUE_POLY] = {
	sizeof(POLY_F3), sizeof(POLY_F4), sizeof(POLY_FT3), sizeof(POLY_FT4),
	sizeof(POLY_G3), sizeof(POLY_G4), sizeof(POLY_GT3), sizeof(POLY_GT4)
};

static const unsigned int divbuf_size[MAX_UNIQUE_POLY] = {
	MAX_DIVBUF_F3, MAX_DIVBUF_F4, MAX_DIVBUF_FT3, MAX_DIVBUF_FT4,
	MAX_DIVBUF_G3, MAX_DIVBUF_G4, MAX_DIVBUF_GT3, MAX_DIVBUF_GT4
};

static POLY_F4 polylight[NUMLIGHTS];
static SVECTOR coord[MAX_VERTS];
static SVECTOR coord_subdiv[MAX_VERTS];
static SVECTOR pointlight[NUMLIGHTS];

/*
 * divbuf_clear:
 *	Reset subdivision buffers.
 */
divbuf_clear(void)
{
	int i;

	for(i = 0; i < MAX_UNIQUE_POLY; i++)
	{
		divbuf.idx[i] = 1;
	}
}

/*
 * divbuf_init:
 *	Initializes all primitives of divbuf.
 */
divbuf_init(void)
{
	int i;

	divbuf.ptr[IDX_F3]	= (char *)divbuf.f3;
	divbuf.ptr[IDX_F4]	= (char *)divbuf.f4;
	divbuf.ptr[IDX_FT3]	= (char *)divbuf.ft3;
	divbuf.ptr[IDX_FT4]	= (char *)divbuf.ft4;
	divbuf.ptr[IDX_G3]	= (char *)divbuf.g3;
	divbuf.ptr[IDX_G4]	= (char *)divbuf.g4;
	divbuf.ptr[IDX_GT3]	= (char *)divbuf.gt3;
	divbuf.ptr[IDX_GT4]	= (char *)divbuf.gt4;

	for(i = 0; i < divbuf_size[IDX_F3]; i++)
		setPolyF3(divbuf.f3+i);

	for(i = 0; i < divbuf_size[IDX_F4]; i++)
		setPolyF4(divbuf.f4+i);

	for(i = 0; i < divbuf_size[IDX_FT3]; i++)
		setPolyFT3(divbuf.ft3+i);

	for(i = 0; i < divbuf_size[IDX_FT4]; i++)
		setPolyFT4(divbuf.ft4+i);

	for(i = 0; i < divbuf_size[IDX_G3]; i++)
		setPolyG3(divbuf.g3+i);

	for(i = 0; i < divbuf_size[IDX_G4]; i++)
		setPolyG4(divbuf.g4+i);

	for(i = 0; i < divbuf_size[IDX_GT3]; i++)
		setPolyGT3(divbuf.gt3+i);

	for(i = 0; i < divbuf_size[IDX_GT4]; i++)
		setPolyGT4(divbuf.gt4+i);

	divbuf.div3.pih = screen_width;
	divbuf.div3.piv = screen_height;
	divbuf.div4.pih = screen_width;
	divbuf.div4.piv = screen_height;
}

/*
 * divbuf_print:
 *	Prints out the primitive index of each divbuf buffer.
 */
divbuf_print(void)
{
	FntPrint("f3 idx: %d\n", divbuf.idx[IDX_F3]);
	FntPrint("f4 idx: %d\n", divbuf.idx[IDX_F4]);
	FntPrint("ft3 idx: %d\n", divbuf.idx[IDX_FT3]);
	FntPrint("ft4 idx: %d\n", divbuf.idx[IDX_FT4]);

	FntPrint("g3 idx: %d\n", divbuf.idx[IDX_G3]);
	FntPrint("g4 idx: %d\n", divbuf.idx[IDX_G4]);
	FntPrint("gt3 idx: %d\n", divbuf.idx[IDX_GT3]);
	FntPrint("gt4 idx: %d\n", divbuf.idx[IDX_GT4]);
}

/*
 * polyclip:
 *	Tests if a polygon should be rendered or not.
 */
static
polyclip(p_sxy, pad, scr_x, scr_y, margin_x, margin_y, nvert)
short *p_sxy;
int pad;
int scr_x, scr_y;
int margin_x, margin_y;
{
	int i, clip = 0;

	for(i = 0; i < nvert; i++)
	{
		if(ABS(scr_x-p_sxy[0]) < scr_x+margin_x)
		{
			if(ABS(scr_y-p_sxy[1]) < scr_y+margin_y)
			{
				clip++;
			}
		}

		p_sxy += pad>>1;
	}

	if(clip < 1)
	{
		return 0;
	}

	return 1;
}

/*
 * process_flag:
 *	Processes common render flag settings. The first
 *	8 bits of the return value are interpreted as a char,
 *	whereas the other 24 bits are treated as a bitmask.
 *	If the char is set to 0, rendering is skipped entirely.
 *
 *	char (1-255):
 *		1: Draw w/ OT
 *		2: Draw w/o OT
 *	bitmask:
 *		bit 8: process lighting for the polygon
 *	
 */
static int
process_flag(flag_runtime, flag_file, p_poly, p_ot, otsize, otz, isomote)
int flag_runtime;
int flag_file;
unsigned long *p_poly, *p_ot, otsize;
long otz, isomote;
{
	int ret = 0;

	if(flag_runtime & XADD_CONVEX_CLIP ||
		(flag_file & XASS_FLAG_CULLING && !(flag_file & XASS_FLAG_BACKFACE_CULL)))
	{
		if(isomote > 0)
		{
			return 0;
		}
	}
	else if(flag_runtime & XADD_CONVEX_CLIP_B ||
		(flag_file & XASS_FLAG_CULLING && flag_file & XASS_FLAG_BACKFACE_CULL))
	{
		if(isomote <= 0)
		{
			return 0;
		}
	}

	if(flag_file & XASS_FLAG_LIGHTING)
	{
		ret |= (1<<8);
	}

	if(flag_runtime & XADD_USE_OT)
	{
		if(otz > MIN_OTZ && otz < otsize)
		{
			ret |= (1<<0);
		}
	}
	else
	{
		ret |= (1<<1);
	}

	return ret;
}

/*
 * subdiv2:
 *	Subdivides a polygon, increments the corresponding
 *	divbuf index, and returns the number of polygons
 *	added to the associated divbuf primitive buffer.
 */
static
subdiv2(v, nvert, p_mtd, p_player_pos, p_camera, p_DIV, p_poly, p_sxy, idx, pad, p_ot_entry)
SVECTOR *v;
unsigned int nvert;
struct AssetMTD *p_mtd;
SVECTOR *p_player_pos;
struct Camera3D *p_camera;
int *p_DIV;
unsigned char *p_poly, *p_sxy;
int idx;
unsigned long pad, *p_ot_entry;
{
	int			i, j;

	CVECTOR			col = {*(p_poly+4), *(p_poly+5), *(p_poly+6), *(p_poly+7)};
	char			*p_suv = p_poly+12;
	char			*p_rgb = p_poly+4;
	int			dist, ndiv = 0;
	unsigned int		dx,dy,dz;
	VECTOR			center, vm[4];

	for(i = 0; i < nvert; i++)
	{
		/* undo translation from vert_srt to get scaled model coords */
		setVector(vm+i, v[i].vx, v[i].vy, v[i].vz);
		vm[i].vx -= (p_mtd->ent.global_pos.vx + p_player_pos->vx) - p_camera->offset.vx;
		vm[i].vy -= (p_mtd->ent.global_pos.vy + p_player_pos->vy) - p_camera->offset.vy;
		vm[i].vz -= (p_mtd->ent.global_pos.vz + p_player_pos->vz) - p_camera->offset.vz;
		setVector(vm+i, -vm[i].vx, -vm[i].vy, -vm[i].vz);
	}
	switch(nvert)
	{
		case 3:
			center.vx = ((vm[0].vx+vm[1].vx+vm[2].vx)*21)>>FIXED_PRECISION;
			center.vy = ((vm[0].vy+vm[1].vy+vm[2].vy)*21)>>FIXED_PRECISION;
			center.vz = ((vm[0].vz+vm[1].vz+vm[2].vz)*21)>>FIXED_PRECISION;
			break;
		case 4:
			center.vx = (vm[0].vx+vm[1].vx+vm[2].vx+vm[3].vx)>>2;
			center.vy = (vm[0].vy+vm[1].vy+vm[2].vy+vm[3].vy)>>2;
			center.vz = (vm[0].vz+vm[1].vz+vm[2].vz+vm[3].vz)>>2;
			break;
	}

//	switch(nvert)
//	{
//		case 3:
//			center.vx = ((v[0].vx+v[1].vx+v[2].vx)*21)>>FIXED_PRECISION;
//			center.vy = ((v[0].vy+v[1].vy+v[2].vy)*21)>>FIXED_PRECISION;
//			center.vz = ((v[0].vz+v[1].vz+v[2].vz)*21)>>FIXED_PRECISION;
//			break;
//		case 4:
//			center.vx = (v[0].vx+v[1].vx+v[2].vx+v[3].vx)>>2;
//			center.vy = (v[0].vy+v[1].vy+v[2].vy+v[3].vy)>>2;
//			center.vz = (v[0].vz+v[1].vz+v[2].vz+v[3].vz)>>2;
//			break;
//	}

	//center.vx <<= 12;
	//center.vy <<= 12;
	//center.vz <<= 12;

	dx = ppow(center.vx-(p_camera->global_pos.vx), 2)>>6;
	dy = ppow(center.vy-(p_camera->global_pos.vy), 2)>>6;
	dz = ppow(center.vz-(p_camera->global_pos.vz), 2)>>6;

	dist = psqrt(dx+dy+dz); /* FIXME: is AABB faster? */

	for(i = 0; i < p_DIV[0]; i++)
	{
		if(!(dist < p_DIV[i+1]))
		{
			if(!i)
			{
				return 0;
			}
			break;
		}
		ndiv++;
	}

	switch(nvert)
	{
		case 3:
			divbuf.div3.cr[0].r0 = &divbuf.div3.r0;
			divbuf.div3.cr[0].r1 = &divbuf.div3.r1;
			divbuf.div3.cr[0].r2 = &divbuf.div3.r2;

			divbuf.div3.r0.v = v[0];
			divbuf.div3.r1.v = v[1];
			divbuf.div3.r2.v = v[2];

			*(unsigned long *)(&divbuf.div3.r0.sxy) = *(unsigned long *)(p_sxy);
			*(unsigned long *)(&divbuf.div3.r1.sxy) = *(unsigned long *)(p_sxy+pad);
			*(unsigned long *)(&divbuf.div3.r2.sxy) = *(unsigned long *)(p_sxy+pad+pad);

			*((unsigned long *)divbuf.div3.r0.uv) = *((unsigned long *)(p_suv));
			*((unsigned long *)divbuf.div3.r1.uv) = *((unsigned long *)(p_suv+pad));
			*((unsigned long *)divbuf.div3.r2.uv) = *((unsigned long *)(p_suv+pad+pad));

			ReadSZfifo3(&divbuf.div3.r0.sz, &divbuf.div3.r1.sz, &divbuf.div3.r2.sz);

			divbuf.div3.ot = p_ot_entry;
			divbuf.div3.rgbc = col;
			divbuf.div3.clut = p_mtd->clut;
			divbuf.div3.tpage = p_mtd->tpage;

			*((unsigned long *)&divbuf.div3.r0.c) = *((unsigned long *)(p_rgb));
			*((unsigned long *)&divbuf.div3.r1.c) = *((unsigned long *)(p_rgb+pad));
			*((unsigned long *)&divbuf.div3.r2.c) = *((unsigned long *)(p_rgb+pad+pad));
			break;
		case 4:
			divbuf.div4.cr[0].r0 = &divbuf.div4.r0;
			divbuf.div4.cr[0].r1 = &divbuf.div4.r1;
			divbuf.div4.cr[0].r2 = &divbuf.div4.r2;
			divbuf.div4.cr[0].r3 = &divbuf.div4.r3;

			divbuf.div4.r0.v = v[0];
			divbuf.div4.r1.v = v[1];
			divbuf.div4.r2.v = v[3];
			divbuf.div4.r3.v = v[2];

			*(unsigned long *)(&divbuf.div4.r0.sxy) = *(unsigned long *)(p_sxy);
			*(unsigned long *)(&divbuf.div4.r1.sxy) = *(unsigned long *)(p_sxy+pad);
			*(unsigned long *)(&divbuf.div4.r2.sxy) = *(unsigned long *)(p_sxy+pad+pad);
			*(unsigned long *)(&divbuf.div4.r3.sxy) = *(unsigned long *)(p_sxy+pad+pad+pad);

			*((unsigned long *)divbuf.div4.r0.uv) = *((unsigned long *)(p_suv));
			*((unsigned long *)divbuf.div4.r1.uv) = *((unsigned long *)(p_suv+pad));
			*((unsigned long *)divbuf.div4.r2.uv) = *((unsigned long *)(p_suv+pad+pad));
			*((unsigned long *)divbuf.div4.r3.uv) = *((unsigned long *)(p_suv+pad+pad+pad));

			ReadSZfifo4(&divbuf.div4.r0.sz, &divbuf.div4.r1.sz, &divbuf.div4.r2.sz, &divbuf.div4.r3.sz);

			divbuf.div4.ot = p_ot_entry;
			divbuf.div4.rgbc = col;
			divbuf.div4.clut = p_mtd->clut;
			divbuf.div4.tpage = p_mtd->tpage;

			*((unsigned long *)&divbuf.div4.r0.c) = *((unsigned long *)(p_rgb));
			*((unsigned long *)&divbuf.div4.r1.c) = *((unsigned long *)(p_rgb+pad));
			*((unsigned long *)&divbuf.div4.r2.c) = *((unsigned long *)(p_rgb+pad+pad));
			*((unsigned long *)&divbuf.div4.r3.c) = *((unsigned long *)(p_rgb+pad+pad+pad));
			break;
	}

	divbuf.div3.ndiv = ndiv;
	divbuf.div4.ndiv = ndiv;

	if(divbuf.idx[idx]+ppow(4, ndiv) > divbuf_size[idx])
	{
		printf("subdiv2() error: subdivision buffers overloaded! (%d/%d); poly:%d\n", divbuf.idx[idx]+ppow(4, ndiv), divbuf_size[idx], idx);
		return 0;
	}

	{
		char		*p_ptr = divbuf.ptr[idx]+divbuf.idx[idx]*apb_size[idx];
		int		npoly = ppow(4, ndiv);

		switch(idx)
		{
			case IDX_F3:
			case IDX_FT3:
			case IDX_G3:
			case IDX_GT3:
				rcptbl[idx](p_ptr, &divbuf.div3);
				break;
			case IDX_F4:
			case IDX_FT4:
			case IDX_G4:
			case IDX_GT4:
				rcptbl[idx](p_ptr, &divbuf.div4);
				break;
		}

		divbuf.idx[idx] += npoly;
	}

	return 1;
}

/*
 * set_sun:
 *	Sets the strength of a given sun.
 */
void
set_sun(idx, x, y, z)
int idx;
short x, y, z;
{
	dgx_light_sun.m[idx][0] = x;
	dgx_light_sun.m[idx][1] = y;
	dgx_light_sun.m[idx][2] = z;
}

/*
 * set_sun_color:
 *	Sets the color of a given sun.
 */
void
set_sun_color(idx, r, g, b)
int idx;
short r, g, b;
{
	dgx_light_sun_color.m[0][idx] = r;
	dgx_light_sun_color.m[1][idx] = g;
	dgx_light_sun_color.m[2][idx] = b;
}

/*
 * vert_srt:
 *	Scales (model-space), rotates (model-space),
 *	and translates (world-space) a vertex.
 *
 *	p_vec (input) Q32.0
 *	p_vert (output) Q10.6
 */
void
vert_srt(p_vec, p_rot, scale, p_world_parent, p_world_player, p_world_camera, flag, p_vert)
VECTOR *p_vec; /* model coord */
MATRIX *p_rot;
unsigned short scale;
SVECTOR *p_world_parent, *p_world_player, *p_world_camera;
int flag;
SVECTOR *p_vert; /* world coord */
{
	if(flag & SRT_S) /* assumes source is from file format e.g. MTD */
	{
		p_vert->vx = (p_vec->vx * scale) >> 12;
		p_vert->vy = (p_vec->vy * scale) >> 12;
		p_vert->vz = (p_vec->vz * scale) >> 12;
	}
	else /* assumes source is a signed short */
	{
		setVector(p_vert, p_vec->vx, p_vec->vy, p_vec->vz);
	}

	if(flag & SRT_R)
	{
		ApplyMatrixSV(p_rot, p_vert, p_vert);
	}

	if(flag & SRT_T)
	{
		p_vert->vx += (p_world_parent->vx + p_world_player->vx) - p_world_camera->vx;
		p_vert->vy += (p_world_parent->vy + p_world_player->vy) - p_world_camera->vy;
		p_vert->vz += (p_world_parent->vz + p_world_player->vz) - p_world_camera->vz;
	}
}

/*
 * light_calc:
 *	Calculates lighting depending on certain flags.
 */
static void
light_calc(p_mtd, p_poly, pad, nvert, poly_idx, v, vn, p_rottrans)
struct AssetMTD *p_mtd;
unsigned long *p_poly;
int pad, nvert, poly_idx;
SVECTOR v[4];
VECTOR *vn;
MATRIX *p_rottrans;
{
	int i, j, k;
	if(p_mtd->flag & XASS_FLAG_PNTLITE)
	{
		if(p_mtd->flag & XASS_FLAG_GOURAUD) /* FIXME: just do one lighting pass on all vertices, then set the correct rgb values for each vertex since there seems to be a LOT of redundancy... */
		{
			int r[4];

			for(j = 0; j < nvert; j++)
			{
				int c;
				int dot;
				VECTOR u = {0, 0, 0};
				VECTOR dist;
				VECTOR normal = {vn[p_mtd->p_indices[poly_idx].n[j]].vx>>FIXED_PRECISION,
					vn[p_mtd->p_indices[poly_idx].n[j]].vy>>FIXED_PRECISION,
					vn[p_mtd->p_indices[poly_idx].n[j]].vz>>FIXED_PRECISION};

				dist.vx = (v[j].vx-pointlight[0].vx);
				dist.vy = (v[j].vy-pointlight[0].vy);
				dist.vz = (v[j].vz-pointlight[0].vz);

				c = psqrt( (ppow(dist.vx, 2)>>FIXED_PRECISION) +
					(ppow(dist.vy, 2)>>FIXED_PRECISION) +
					(ppow(dist.vz, 2)>>FIXED_PRECISION) );

				{
					int tmp = c>>FIXED_PRECISION;

					if(tmp != 0)
					{
						u.vx = dist.vx/tmp;
						u.vy = dist.vy/tmp;
						u.vz = dist.vz/tmp;
					}
				}

				dot = pdot(&u, &normal)/3;

				r[j] = (dgx_light_point_radius[0]/c*dot)>>FIXED_PRECISION;
				r[j] = ABS(pclamp(r[j], 0, 255));

				{
					unsigned int tmp_r = (dgx_light_point_color[0].r*r[j])>>8;
					unsigned int tmp_g = (dgx_light_point_color[0].g*r[j])>>8;
					unsigned int tmp_b = (dgx_light_point_color[0].b*r[j])>>8;

					if(j == 3) /* ensure 0132 rgb for quads */
					{
						*(((unsigned char *)p_poly)+4+(j*pad)+0) = *(((unsigned char *)p_poly)+4+((j-1)*pad)+0);
						*(((unsigned char *)p_poly)+4+(j*pad)+1) = *(((unsigned char *)p_poly)+4+((j-1)*pad)+1);
						*(((unsigned char *)p_poly)+4+(j*pad)+2) = *(((unsigned char *)p_poly)+4+((j-1)*pad)+2);

						*(((unsigned char *)p_poly)+4+((j-1)*pad)+0) = tmp_r;
						*(((unsigned char *)p_poly)+4+((j-1)*pad)+1) = tmp_g;
						*(((unsigned char *)p_poly)+4+((j-1)*pad)+2) = tmp_b;
					}
					else
					{
						*(((unsigned char *)p_poly)+4+(j*pad)+0) = tmp_r;
						*(((unsigned char *)p_poly)+4+(j*pad)+1) = tmp_g;
						*(((unsigned char *)p_poly)+4+(j*pad)+2) = tmp_b;
					}
				}
			}
		}
		else /* flat-shaded point light */
		{
			int r, c, dot;
			SVECTOR f;
			VECTOR u = {0, 0, 0};
			VECTOR dist;
			VECTOR normal = {vn[p_mtd->p_indices[poly_idx].n[0]].vx>>FIXED_PRECISION,
				vn[p_mtd->p_indices[poly_idx].n[0]].vy>>FIXED_PRECISION,
				vn[p_mtd->p_indices[poly_idx].n[0]].vz>>FIXED_PRECISION};

			{
				VECTOR vec = {0, 0, 0};
				for(j = 0; j < nvert; j++)
				{
					vec.vx += v[j].vx;
					vec.vy += v[j].vy;
					vec.vz += v[j].vz;
				}
				setVector(&f, vec.vx/nvert, vec.vy/nvert, vec.vz/nvert);
			}

			dist.vx = (f.vx-pointlight[0].vx);
			dist.vy = (f.vy-pointlight[0].vy);
			dist.vz = (f.vz-pointlight[0].vz);

			c = psqrt( (ppow(dist.vx, 2)>>FIXED_PRECISION) +
				(ppow(dist.vy, 2)>>FIXED_PRECISION) +
				(ppow(dist.vz, 2)>>FIXED_PRECISION) );

			{
				int tmp = c>>FIXED_PRECISION;

				if(tmp != 0)
				{
					u.vx = dist.vx/tmp;
					u.vy = dist.vy/tmp;
					u.vz = dist.vz/tmp;
				}
			}

			dot = pdot(&u, &normal)/3;

			r = (dgx_light_point_radius[0]/c*dot)>>FIXED_PRECISION;
			r = ABS(pclamp(r, 0, 255));

			{
				unsigned int tmp_r = (dgx_light_point_color[0].r*r)>>8;
				unsigned int tmp_g = (dgx_light_point_color[0].g*r)>>8;
				unsigned int tmp_b = (dgx_light_point_color[0].b*r)>>8;

				*(((unsigned char *)p_poly)+4) = tmp_r;
				*(((unsigned char *)p_poly)+5) = tmp_g;
				*(((unsigned char *)p_poly)+6) = tmp_b;
			}
		}
	}
	else /* directional */
	{
		if(p_mtd->flag & XASS_FLAG_GOURAUD) /* FIXME: just do one lighting pass on all vertices, then set the correct rgb values for each vertex since there seems to be a LOT of redundancy... */
		{
			for(j = 0; j < nvert; j++)
			{
				int r, c, dot;
				VECTOR u = {dgx_light_sun_ang[0].vx, dgx_light_sun_ang[0].vy, dgx_light_sun_ang[0].vz};
				VECTOR normal = {vn[p_mtd->p_indices[poly_idx].n[j]].vx>>FIXED_PRECISION,
					vn[p_mtd->p_indices[poly_idx].n[j]].vy>>FIXED_PRECISION,
					vn[p_mtd->p_indices[poly_idx].n[j]].vz>>FIXED_PRECISION};

				dot = pdot(&u, &normal)/3;

				r = ((dgx_light_point_radius[0]>>17)*dot)>>FIXED_PRECISION;
				r = ABS(pclamp(r, 0, 255));

				{
					unsigned int tmp_r = (dgx_light_point_color[0].r*r)>>8;
					unsigned int tmp_g = (dgx_light_point_color[0].g*r)>>8;
					unsigned int tmp_b = (dgx_light_point_color[0].b*r)>>8;

					if(j == 3) /* ensure 0132 rgb for quads */
					{
						*(((unsigned char *)p_poly)+4+(j*pad)+0) = *(((unsigned char *)p_poly)+4+((j-1)*pad)+0);
						*(((unsigned char *)p_poly)+4+(j*pad)+1) = *(((unsigned char *)p_poly)+4+((j-1)*pad)+1);
						*(((unsigned char *)p_poly)+4+(j*pad)+2) = *(((unsigned char *)p_poly)+4+((j-1)*pad)+2);

						*(((unsigned char *)p_poly)+4+((j-1)*pad)+0) = tmp_r;
						*(((unsigned char *)p_poly)+4+((j-1)*pad)+1) = tmp_g;
						*(((unsigned char *)p_poly)+4+((j-1)*pad)+2) = tmp_b;
					}
					else
					{
						*(((unsigned char *)p_poly)+4+(j*pad)+0) = tmp_r;
						*(((unsigned char *)p_poly)+4+(j*pad)+1) = tmp_g;
						*(((unsigned char *)p_poly)+4+(j*pad)+2) = tmp_b;
					}
				}
			}
		}
		else
		{
			int r, dot;
			VECTOR u = {dgx_light_sun_ang[0].vx, dgx_light_sun_ang[0].vy, dgx_light_sun_ang[0].vz};
			VECTOR normal = {vn[p_mtd->p_indices[poly_idx].n[0]].vx>>FIXED_PRECISION,
				vn[p_mtd->p_indices[poly_idx].n[0]].vy>>FIXED_PRECISION,
				vn[p_mtd->p_indices[poly_idx].n[0]].vz>>FIXED_PRECISION};

			dot = pdot(&u, &normal)/3;

			r = ((dgx_light_point_radius[0]>>17)*dot)>>FIXED_PRECISION;
			r = ABS(pclamp(r, 0, 255));

			{
				unsigned int tmp_r = (dgx_light_point_color[0].r*r)>>8;
				unsigned int tmp_g = (dgx_light_point_color[0].g*r)>>8;
				unsigned int tmp_b = (dgx_light_point_color[0].b*r)>>8;

				*(((unsigned char *)p_poly)+4) = tmp_r;
				*(((unsigned char *)p_poly)+5) = tmp_g;
				*(((unsigned char *)p_poly)+6) = tmp_b;
			}
		}
	}
}

/*
 * mtd_add:
 *	Adds the primitives associated with an MTD file
 *	into an ordering table.
 */
void
mtd_add(p_mtd, p_player_pos, p_rottrans, rotm, p_camera, flag, p_ot, otsize, p_DIV)
struct AssetMTD *p_mtd;
SVECTOR *p_player_pos;
MATRIX *p_rottrans;
int rotm; /* model-space rotation order */
struct Camera3D *p_camera;
int flag, *p_ot, otsize, *p_DIV;
{
	int			i, j, k;
	unsigned short		ridx[MAX_UNIQUE_POLY] = {
		0, 0, 0, 0,
		0, 0, 0, 0
	};

	long			otz, p, flg, isomote;
	SVECTOR			*v  = 0x1f800000;
	SVECTOR			*vs = 0x1f800008;
	VECTOR			*vp = p_mtd->p_vertex;
	VECTOR			*vn = p_mtd->p_normal;

	int			ndiv;
	void			*p_apb = (void *)&p_mtd->apb;

	const int		USE_WIREFRAME	= p_mtd->flag & XASS_FLAG_WIREFRAME;
	const int		USE_TEXTURE	= p_mtd->flag & XASS_FLAG_TEXTURED;
	const int		USE_GOURAUD	= p_mtd->flag & XASS_FLAG_GOURAUD;
	const int		USE_LIGHTING	= p_mtd->flag & XASS_FLAG_LIGHTING;
	const int		USE_PNTLITE	= p_mtd->flag & XASS_FLAG_PNTLITE;
	MATRIX			model_rot[2];
	VECTOR			mid = {0, 0, 0};

	PushMatrix();

	rotmat[rotm](&p_mtd->ent.rotation_model, model_rot);

	RotMatrix_gte(&p_mtd->ent.rotation, p_rottrans);
	TransMatrix(p_rottrans, &p_mtd->ent.translation);

	/* derive world coords all at once */
	{
		int FLAG_SRT = ((flag&0xe00000)>>21)^7; /* check for bits that disable SRT */

		if(flag & XADD_USE_PIVOT_LOCAL2)
		{
			VECTOR tmp;

			for(i = 0; i < p_mtd->nvert; i++)
			{
				mid.vx += (vp[i].vx);
				mid.vy += (vp[i].vy);
				mid.vz += (vp[i].vz);
			}

			mid.vx /= p_mtd->nvert;
			mid.vy /= p_mtd->nvert;
			mid.vz /= p_mtd->nvert;

			rotmat[p_mtd->rotm2](&p_mtd->rotmod2, model_rot+1);

			for(i = 0; i < p_mtd->nvert; i++)
			{
				vp[i].vx -= mid.vx;
				vp[i].vy -= mid.vy;
				vp[i].vz -= mid.vz;

				ApplyMatrixLV(model_rot+1, &vp[i], &tmp);

				vp[i].vx += mid.vx;
				vp[i].vy += mid.vy;
				vp[i].vz += mid.vz;

				tmp.vx += mid.vx;
				tmp.vy += mid.vy;
				tmp.vz += mid.vz;

				vert_srt(&tmp, model_rot, p_mtd->ent.scale, &p_mtd->ent.global_pos,
					p_player_pos, &p_camera->offset, FLAG_SRT, &coord[i]);
			}
		}
		else if(flag&XADD_USE_PIVOT_LOCAL3) /* XADD_USE_PIVOT_LOCAL3 is rather slow */
		{
			rotmat[p_mtd->rotm2](&p_mtd->rotmod2, model_rot+1);

			for(i = 0; i < p_mtd->nvert; i++)
			{
				ApplyMatrixLV(model_rot+1, &vp[i], &vp[i]);
				vert_srt(&vp[i], model_rot, p_mtd->ent.scale, &p_mtd->ent.global_pos,
						p_player_pos, &p_camera->offset, FLAG_SRT, &coord[i]);
			}
		}
		else
		{
			/* we store the results of scale and rotation for later use in subdiv2() */
			if(flag&XADD_SUBDIV)
			{
				//VECTOR tmp;
				for(i = 0; i < p_mtd->nvert; i++)
				{
					//vert_srt(&vp[i], model_rot, p_mtd->ent.scale, &p_mtd->ent.global_pos,
					//		p_player_pos, &p_camera->offset, FLAG_SRT&~SRT_T, &coord_subdiv[i]);
					//setVector(&tmp, coord_subdiv[i].vx, coord_subdiv[i].vy, coord_subdiv[i].vz);
					/* IMPORTANT: it is assumed that XADD_NO_SRT_S and XADD_NO_SRT_R is passed here */
					vert_srt(&vp[i], model_rot, p_mtd->ent.scale, &p_mtd->ent.global_pos,
							p_player_pos, &p_camera->offset, SRT_T, &coord[i]);
				}
			}
			else
			{
				for(i = 0; i < p_mtd->nvert; i++)
				{
					vert_srt(&vp[i], model_rot, p_mtd->ent.scale, &p_mtd->ent.global_pos,
							p_player_pos, &p_camera->offset, FLAG_SRT, &coord[i]);
				}
			}

		}
	}

	gte_SetRotMatrix(p_rottrans);
	gte_SetTransMatrix(p_rottrans);

	for(i = 0; i < p_mtd->npoly; i++)
	{
		int		r = 0;
		unsigned int	polytype = p_mtd->p_indices[i].type;

		if(USE_WIREFRAME)
		{
			otz =
			RotTransPers(&coord[p_mtd->p_indices[i].v[0]], (long *)&p_mtd->apb.p_linef2[i].x0, &p, &flg);
			RotTransPers(&coord[p_mtd->p_indices[i].v[1]], (long *)&p_mtd->apb.p_linef2[i].x1, &p, &flg);

			if(flag & XADD_USE_OT)
			{
				if(otz > MIN_OTZ && otz < otsize)
				{
					AddPrim(p_ot+otz, &p_mtd->apb.p_linef2[i]);
				}
			}
			else
			{
				DrawPrim(&p_mtd->apb.p_linef2[i]);
			}
		}
		else
		{
			unsigned long *p_poly = p_apb;
			int pad = 4;
			int nvert = !(polytype%2) ? 3 : 4;

			for(j = 0; j < 4; j++)
			{
				v[j] = coord[p_mtd->p_indices[i].v[j]];
				setVector(vs+j, vp[p_mtd->p_indices[i].v[j]].vx, vp[p_mtd->p_indices[i].v[j]].vy, vp[p_mtd->p_indices[i].v[j]].vz);
			}

			if(USE_GOURAUD)
			{
				r+=4;
				pad+=4;
			}
			if(USE_TEXTURE)
			{
				r+=2;
				pad+=4;
			}
			if(polytype)
			{
				r+=1;
			}

			/* select and dereference buffer from apb */
			p_poly += r;
			p_poly = ((char *)(*p_poly))+apb_size[r]*ridx[r];

			otz = rottbl[polytype](v, flag, pad, p_poly+2, &isomote, NULL);

			if(flag&XADD_SAVE_OTZ)
			{
				if(p_mtd->npoly > MAX_OTZBUF)
				{
					printf("mtd_add() error: cannot save otz values: (%d/%d) polygons. XADD_SAVE_OTZ flag must be turned off!\n", p_mtd->npoly, MAX_OTZBUF);
					while(1);
				}
				saveotz[i] = otz;
				saveotz_type[i] = r;
			}

			if((flag & XADD_SUBDIV) && (flag & XADD_USE_OT))
			{
				if(otz > MIN_OTZ && otz < otsize)
				{
					long otz2;

#ifdef WOBBLE_SUBDIV
					VECTOR lvec[4];
					SVECTOR svec[4];
					VECTOR mid;

					switch(nvert)
					{
						case 3:
							mid.vx = ((v[0].vx+v[1].vx+v[2].vx)*21)>>FIXED_PRECISION;
							mid.vy = ((v[0].vy+v[1].vy+v[2].vy)*21)>>FIXED_PRECISION;
							mid.vz = ((v[0].vz+v[1].vz+v[2].vz)*21)>>FIXED_PRECISION;
							break;
						case 4:
							mid.vx = (v[0].vx+v[1].vx+v[2].vx+v[3].vx)>>2;
							mid.vy = (v[0].vy+v[1].vy+v[2].vy+v[3].vy)>>2;
							mid.vz = (v[0].vz+v[1].vz+v[2].vz+v[3].vz)>>2;
							break;
					}

					/*
					 * Move primitive to its own model-space, enlarge it *slightly*, then
					 * translate the primitive back to its parent model's model-space. The 
					 * slightly enlarged version of the original primitive is passed onto
					 * the subdivision routine, effectively eliminating (most) seams. Although
					 * this method does not require any additional polygons, the resulting
					 * model may be a bit jank...
					 */
					for(j = 0; j < nvert; j++)
					{
						setVector(lvec+j, v[j].vx, v[j].vy, v[j].vz);

						lvec[j].vx -= mid.vx;
						lvec[j].vy -= mid.vy;
						lvec[j].vz -= mid.vz;

						/* 106.25% (in 26.6) */
						lvec[j].vx *= 68;
						lvec[j].vy *= 68;
						lvec[j].vz *= 68;

						lvec[j].vx >>= 6;
						lvec[j].vy >>= 6;
						lvec[j].vz >>= 6;

						lvec[j].vx += mid.vx;
						lvec[j].vy += mid.vy;
						lvec[j].vz += mid.vz;
						setVector(svec+j, lvec[j].vx, lvec[j].vy, lvec[j].vz);
					}
					otz2 = rottbl[polytype](svec, flag, pad, divbuf.ptr[r]+8, &isomote, NULL);

					if( subdiv2(svec, nvert, p_mtd, p_player_pos, p_camera, p_DIV, p_poly, divbuf.ptr[r]+8, r, pad, p_ot+otz2) )
					{
						goto RENDER_SKIP;
					}
#else
					otz2 = rottbl[polytype](v, flag, pad, divbuf.ptr[r]+8, &isomote, NULL);

					if( subdiv2(v, nvert, p_mtd, p_player_pos, p_camera, p_DIV, p_poly, divbuf.ptr[r]+8, r, pad, p_ot+otz2) )
					//if( subdiv2(v, nvert, p_mtd, p_player_pos, p_camera, p_DIV, p_poly, p_poly+2, r, pad, p_ot+otz) )
					{
						goto RENDER_SKIP;
					}
#endif/* WOBBLE_SUBDIV */
				}
				else
				{
					if(flag&XADD_SAVE_OTZ)
					{
						saveotz_is_visble[i] = 0;
					}
					goto RENDER_SKIP;
				}
			}

			if( polyclip(p_poly+2, pad, screen_width>>1, screen_height>>1, 16, 16, nvert) )
			{
				int val = process_flag(flag, p_mtd->flag, p_poly, p_ot, otsize, otz, isomote);

				if((val & (1<<8)) /* && (val & 0xff)*/)
				{
					/*
					 * dgx_light_point_update() should be called first before mtd_add()
					 * in your game loop.
					 */
					light_calc(p_mtd, p_poly, pad, nvert, i, v, vn, p_rottrans);
				}

				if(flag&XADD_SAVE_OTZ)
				{
					saveotz_is_visble[i] = val&0xff;
				}

				switch(val & 0xff)
				{
					case 0:
						goto RENDER_SKIP;
					case 1:
						AddPrim(p_ot+otz, p_poly);
						break;
					case 2:
						DrawPrim(p_poly);
						break;
					default:
						printf("mtd_add -> process_flag: Invalid bit setting - (0x%x)\n", val);
						break;
				}
			}
RENDER_SKIP:
			if(ridx[r] < USHRT_MAX)
			{
				ridx[r]++;
			}
		}
	}

	PopMatrix();
}

/*
 * atd_add:
 *	Adds the primitives associated with an ATD file
 *	into an ordering table.
 */
void
atd_add(p_atd, p_player_pos, p_rottrans, rotm, p_camera, flag, p_ot, otsize, p_DIV)
struct AssetATD *p_atd;
SVECTOR *p_player_pos;
MATRIX *p_rottrans;
int rotm; /* model-space rotation order */
struct Camera3D *p_camera;
int flag, *p_ot, otsize, *p_DIV;
{
	int			i, j, k;
	unsigned short		ridx[MAX_UNIQUE_POLY] = {
		0, 0, 0, 0,
		0, 0, 0, 0
	};

	long			otz, p, flg, isomote;
	SVECTOR			v[4];
	struct ATD_VERTEX	*vp = p_atd->p_vertex;
	VECTOR			*vn = p_atd->p_normal;

	int			ndiv;
	void			*p_apb = (void *)&p_atd->apb;

	const int		USE_WIREFRAME	= p_atd->flag & XASS_FLAG_WIREFRAME;
	const int		USE_TEXTURE	= p_atd->flag & XASS_FLAG_TEXTURED;
	const int		USE_GOURAUD	= p_atd->flag & XASS_FLAG_GOURAUD;
	const int		USE_LIGHTING	= p_atd->flag & XASS_FLAG_LIGHTING;
	const int		USE_PNTLITE	= p_atd->flag & XASS_FLAG_PNTLITE;
	MATRIX			model_rot[2];
	VECTOR			mid = {0, 0, 0};

	PushMatrix();

	rotmat[rotm](&p_atd->ent.rotation_model, model_rot);

	RotMatrix_gte(&p_atd->ent.rotation, p_rottrans);
	TransMatrix(p_rottrans, &p_atd->ent.translation);

	switch(p_atd->intr)
	{
		case INSTANT:
ATD_INTR_INSTANT:
		{
			int FLAG_SRT = ((flag&0xe00000)>>21)^7; /* check for bits that disable SRT */
			for(i = 0; i < p_atd->nvert; i++)
			{
				/* derive world coords all at once */
				for(i = 0; i < p_atd->nvert; i++)
				{
					vert_srt(&vp[i].p_keyframe_vector[p_atd->displayed_keyframe], model_rot, p_atd->ent.scale, &p_atd->ent.global_pos,
							p_player_pos, &p_camera->offset, FLAG_SRT, &coord[i]);
				}
			}
		}
			break;
		case LINEAR:
			if( !(p_atd->displayed_keyframe < p_atd->nkey-1) )
			{
				goto ATD_INTR_INSTANT;
			}
			else
			{
				/*
				 * since every vertex has the same number of keyframes at the
				 * same times, we can just use the zeroth vertex.
				 */
				int k0		= vp[0].p_keyframe_framenum[p_atd->displayed_keyframe];		/* begin */
				int k1		= vp[0].p_keyframe_framenum[p_atd->displayed_keyframe+1];		/* end */
				int cursor	= p_atd->timeline_cursor;
				unsigned int t;

				/* normalize k1 end point */
				k1 -= k0;

				/* adjust cursor to fit within 0..k1 line */
				cursor -= k0;

				/*
				 * where is our cursor relative to k0 and k1?
				 * (percentage-wise, k0 being 0% and k1 being 100%)
				 */
				cursor <<= 12;
				t = cursor / k1;

				/* FIXME: use lerpi() from math.c instead. */
				for(i = 0; i < p_atd->nvert; i++)
				{
					VECTOR v0 = vp[i].p_keyframe_vector[p_atd->displayed_keyframe];
					VECTOR v1 = vp[i].p_keyframe_vector[p_atd->displayed_keyframe+1];

					/* lerp! */
					if(!(flag & XADD_NO_SRT_S))
					{
						coord[i].vx = ((((4096-t)*v0.vx + v1.vx*t)>>12) * p_atd->ent.scale)>>12;
						coord[i].vy = ((((4096-t)*v0.vy + v1.vy*t)>>12) * p_atd->ent.scale)>>12;
						coord[i].vz = ((((4096-t)*v0.vz + v1.vz*t)>>12) * p_atd->ent.scale)>>12;
					}
					else
					{
						coord[i].vx = ((4096-t)*v0.vx + v1.vx*t)>>12;
						coord[i].vy = ((4096-t)*v0.vy + v1.vy*t)>>12;
						coord[i].vz = ((4096-t)*v0.vz + v1.vz*t)>>12;
					}

					if(!(flag & XADD_NO_SRT_R))
					{
						/* rotate in model space */
						ApplyMatrixSV(model_rot, &coord[i], &coord[i]);
					}

					if(!(flag & XADD_NO_SRT_T))
					{
						coord[i].vx += (p_atd->ent.global_pos.vx + p_player_pos->vx) - p_camera->offset.vx;
						coord[i].vy += (p_atd->ent.global_pos.vy + p_player_pos->vy) - p_camera->offset.vy;
						coord[i].vz += (p_atd->ent.global_pos.vz + p_player_pos->vz) - p_camera->offset.vz;
					}
				}
			}
			break;
		default:
			printf("atd_add() error: unknown interpolation type! (%d)\n", p_atd->intr);
			while(1);
	}


	gte_SetRotMatrix(p_rottrans);
	gte_SetTransMatrix(p_rottrans);

	for(i = 0; i < p_atd->npoly; i++)
	{
		int		r = 0;
		unsigned int	polytype = p_atd->p_indices[i].type;

		if(USE_WIREFRAME)
		{
			otz =
			RotTransPers(&coord[p_atd->p_indices[i].v[0]], (long *)&p_atd->apb.p_linef2[i].x0, &p, &flg);
			RotTransPers(&coord[p_atd->p_indices[i].v[1]], (long *)&p_atd->apb.p_linef2[i].x1, &p, &flg);

			if(flag & XADD_USE_OT)
			{
				if(otz > MIN_OTZ && otz < otsize)
				{
					AddPrim(p_ot+otz, &p_atd->apb.p_linef2[i]);
				}
			}
			else
			{
				DrawPrim(&p_atd->apb.p_linef2[i]);
			}
		}
		else
		{
			unsigned long *p_poly = p_apb;
			int pad = 4;
			int nvert = !(polytype%2) ? 3 : 4;

			for(j = 0; j < 4; j++)
			{
				v[j] = coord[p_atd->p_indices[i].v[j]];
			}

			if(USE_GOURAUD)
			{
				r+=4;
				pad+=4;
			}
			if(USE_TEXTURE)
			{
				r+=2;
				pad+=4;
			}
			if(polytype)
			{
				r+=1;
			}

			/* select and dereference buffer from apb */
			p_poly += r;
			p_poly = ((char *)(*p_poly))+apb_size[r]*ridx[r];

			otz = rottbl[polytype](v, flag, pad, p_poly+2, &isomote, NULL);

			if(flag&XADD_SAVE_OTZ)
			{
				if(p_atd->npoly > MAX_OTZBUF)
				{
					printf("atd_add() error: cannot save otz values: (%d/%d) polygons. XADD_SAVE_OTZ flag must be turned off!\n", p_atd->npoly, MAX_OTZBUF);
					while(1);
				}
				saveotz[i] = otz;
				saveotz_type[i] = r;
			}

			if((flag & XADD_SUBDIV) && (flag & XADD_USE_OT))
			{
				if(otz > MIN_OTZ && otz < otsize)
				{
					VECTOR lvec[4];
					SVECTOR svec[4];
					VECTOR mid;
					long otz2;

					switch(nvert)
					{
						case 3:
							mid.vx = ((v[0].vx+v[1].vx+v[2].vx)*21)>>FIXED_PRECISION;
							mid.vy = ((v[0].vy+v[1].vy+v[2].vy)*21)>>FIXED_PRECISION;
							mid.vz = ((v[0].vz+v[1].vz+v[2].vz)*21)>>FIXED_PRECISION;
							break;
						case 4:
							mid.vx = (v[0].vx+v[1].vx+v[2].vx+v[3].vx)>>2;
							mid.vy = (v[0].vy+v[1].vy+v[2].vy+v[3].vy)>>2;
							mid.vz = (v[0].vz+v[1].vz+v[2].vz+v[3].vz)>>2;
							break;
					}

					/*
					 * Move primitive to its own model-space, enlarge it *slightly*, then
					 * translate the primitive back to its parent model's model-space. The 
					 * slightly enlarged version of the original primitive is passed onto
					 * the subdivision routine, effectively eliminating (most) seams. Although
					 * this method does not require any additional polygons, the resulting
					 * model may be a bit jank...
					 */
					for(j = 0; j < nvert; j++)
					{
						setVector(lvec+j, v[j].vx, v[j].vy, v[j].vz);

						lvec[j].vx -= mid.vx;
						lvec[j].vy -= mid.vy;
						lvec[j].vz -= mid.vz;

						/* 106.25% (in 26.6) */
						lvec[j].vx *= 68;
						lvec[j].vy *= 68;
						lvec[j].vz *= 68;

						lvec[j].vx >>= 6;
						lvec[j].vy >>= 6;
						lvec[j].vz >>= 6;

						lvec[j].vx += mid.vx;
						lvec[j].vy += mid.vy;
						lvec[j].vz += mid.vz;
						setVector(svec+j, lvec[j].vx, lvec[j].vy, lvec[j].vz);
					}

					otz2 = rottbl[polytype](v, flag, pad, divbuf.ptr[r]+8, &isomote, NULL);

					if( subdiv2(svec, nvert, p_atd, p_player_pos, p_camera, p_DIV, p_poly, divbuf.ptr[r]+8, r, pad, p_ot+otz2) )
					{
						goto RENDER_SKIP;
					}
				}
				else
				{
					if(flag&XADD_SAVE_OTZ)
					{
						saveotz_is_visble[i] = 0;
					}
					goto RENDER_SKIP;
				}
			}

			if( polyclip(p_poly+2, pad, screen_width>>1, screen_height>>1, 16, 16, nvert) )
			{
				int val = process_flag(flag, p_atd->flag, p_poly, p_ot, otsize, otz, isomote);

				if((val & (1<<8)) /* && (val & 0xff)*/)
				{
					/*
					 * dgx_light_point_update() should be called first before atd_add()
					 * in your game loop.
					 */
					light_calc(p_atd, p_poly, pad, nvert, i, v, vn, p_rottrans);
				}

				if(flag&XADD_SAVE_OTZ)
				{
					saveotz_is_visble[i] = val&0xff;
				}

				switch(val & 0xff)
				{
					case 0:
						goto RENDER_SKIP;
					case 1:
						AddPrim(p_ot+otz, p_poly);
						break;
					case 2:
						DrawPrim(p_poly);
						break;
					default:
						printf("atd_add -> process_flag: Invalid bit setting - (0x%x)\n", val);
						break;
				}
			}
RENDER_SKIP:
			if(ridx[r] < USHRT_MAX)
			{
				ridx[r]++;
			}
		}
	}

	PopMatrix();
}

#define DEREF_PARENT_ENT(_p) \
	p_parent_rotation		= &(_p)->ent.rotation, \
	p_parent_rotation_model		= &(_p)->ent.rotation_model, \
	p_parent_translation		= &(_p)->ent.translation, \
	p_parent_global_pos		= &(_p)->ent.global_pos, \
	p_parent_scale			= &(_p)->ent.scale

/*
 * cxc_add:
 *	Adds the primitives associated with an CXC file
 *	into an ordering table.
 *
 *	FIXME: ANCIENT FUNCTION: USE ONLY IF DEBUGGING
 */
cxc_add(p_ot, p_cxc, p_rottrans, p_global_pos, p_camera, use_ot, use_pivot_local, otsize)
unsigned long *p_ot;
struct AssetCXC *p_cxc;
MATRIX *p_rottrans;
SVECTOR *p_global_pos;
struct Camera3D *p_camera;
int use_ot;
int use_pivot_local;
int otsize;
{
	int			i, j = 0, k;
	long			otz, p, flg;
	SVECTOR			v[4];

	SVECTOR			*p_parent_rotation;
	SVECTOR			*p_parent_rotation_model;
	VECTOR			*p_parent_translation;
	SVECTOR			*p_parent_global_pos;
	unsigned short		*p_parent_scale;

	MATRIX			model_rot;

	/* determine how to dereference parent */
	switch(p_cxc->parent)
	{
		case CXC_PARENT_NONE:
		{
			struct AssetCXC *p = p_cxc;
			DEREF_PARENT_ENT(p);
			break;
		}
		case CXC_PARENT_MTD:
		{
			struct AssetMTD *p = (struct AssetMTD *)p_cxc->p_parent;
			DEREF_PARENT_ENT(p);
			break;
		}
		case CXC_PARENT_ATD:
		{
			struct AssetATD *p = (struct AssetATD *)p_cxc->p_parent;
			DEREF_PARENT_ENT(p);
			break;
		}
		default:
			printf("cxc_add() error: unknown CXC parent (%d) !\n", p_cxc->parent);
			while(1);
	}

	PushMatrix();

	RotMatrix(p_parent_rotation, p_rottrans);
	TransMatrix(p_rottrans, p_parent_translation);

	SetRotMatrix(p_rottrans);
	SetTransMatrix(p_rottrans);

	/* render COLMAP's quads */
	for(i = 0; i < p_cxc->COLMAP.nquads; i++)
	{
		/* FIXME: use vert_srt */
		for(j = 0; j < 4; j++)
		{
			/* scale */
			v[j].vx = p_cxc->COLMAP.p_quad[i].global_coords.v[j].vx * (*p_parent_scale);
			v[j].vy = p_cxc->COLMAP.p_quad[i].global_coords.v[j].vy * (*p_parent_scale);
			v[j].vz = p_cxc->COLMAP.p_quad[i].global_coords.v[j].vz * (*p_parent_scale);

			/* rotate in model space */
			if(use_pivot_local)
			{
				RotMatrix(p_parent_rotation_model, &model_rot);
				ApplyMatrixSV(&model_rot, &v[j], &v[j]);
			}
			/* translate */
			v[j].vx += p_parent_global_pos->vx + p_global_pos->vx - p_camera->offset.vx;
			v[j].vy += p_parent_global_pos->vy + p_global_pos->vy - p_camera->offset.vy;
			v[j].vz += p_parent_global_pos->vz + p_global_pos->vz - p_camera->offset.vz;
		}


		otz = RotAverage4(&v[0], &v[1], &v[2], &v[3],
			(long *)&p_cxc->COLMAP.p_quad[i].prim[0].x0,
			(long *)&p_cxc->COLMAP.p_quad[i].prim[1].x0,
			(long *)&p_cxc->COLMAP.p_quad[i].prim[3].x0,
			(long *)&p_cxc->COLMAP.p_quad[i].prim[2].x0,
			&p, &flg);

		///* FIXME: DEBUG shit */
		//{
		//	SVECTOR vd[2];
		//	long dmy;

		//	for(j = 0; j < 2; j++)
		//	{
		//		setVector(&vd[j], p_cxc->COLMAP.p_quad[i].aabb_x[j], p_cxc->COLMAP.p_quad[i].aabb_y[j], p_cxc->COLMAP.p_quad[i].aabb_z[j]);

		//		vd[j].vx *= *p_parent_scale;
		//		vd[j].vy *= *p_parent_scale;
		//		vd[j].vz *= *p_parent_scale;

		//		vd[j].vx += p_parent_global_pos->vx + p_global_pos->vx - p_camera->offset.vx;
		//		vd[j].vy += p_parent_global_pos->vy + p_global_pos->vy - p_camera->offset.vy;
		//		vd[j].vz += p_parent_global_pos->vz + p_global_pos->vz - p_camera->offset.vz;
		//	}

		//	RotAverage3(&vd[1], &vd[0], &vd[1],
		//		(long *)&p_cxc->COLMAP.p_quad[i].aabb_corner[0].x0,
		//		(long *)&p_cxc->COLMAP.p_quad[i].aabb_corner[1].x0,
		//		&dmy, &p, &flg);

		//	/* FIXME: DEBUG */
		//	if(i == 6)
		//	{
		//		DrawPrim(&p_cxc->COLMAP.p_quad[i].aabb_corner[0]);
		//		DrawPrim(&p_cxc->COLMAP.p_quad[i].aabb_corner[1]);
		//	}
		//}

		/* connect LINE_F2s */
		p_cxc->COLMAP.p_quad[i].prim[0].x1 = p_cxc->COLMAP.p_quad[i].prim[2].x0;
		p_cxc->COLMAP.p_quad[i].prim[1].x1 = p_cxc->COLMAP.p_quad[i].prim[0].x0;
		p_cxc->COLMAP.p_quad[i].prim[3].x1 = p_cxc->COLMAP.p_quad[i].prim[1].x0;
		p_cxc->COLMAP.p_quad[i].prim[2].x1 = p_cxc->COLMAP.p_quad[i].prim[3].x0;

		p_cxc->COLMAP.p_quad[i].prim[0].y1 = p_cxc->COLMAP.p_quad[i].prim[2].y0;
		p_cxc->COLMAP.p_quad[i].prim[1].y1 = p_cxc->COLMAP.p_quad[i].prim[0].y0;
		p_cxc->COLMAP.p_quad[i].prim[3].y1 = p_cxc->COLMAP.p_quad[i].prim[1].y0;
		p_cxc->COLMAP.p_quad[i].prim[2].y1 = p_cxc->COLMAP.p_quad[i].prim[3].y0;

		/* either add to OT or not */
		if( use_ot && (otz > MIN_OTZ) && (otz < otsize) )
		{
			for(j = 0; j < 4; j++)
			{
				AddPrim(p_ot+otz, &p_cxc->COLMAP.p_quad[i].prim[j]);
			}
		}
		else
		{
			for(j = 0; j < 4; j++)
			{
				DrawPrim(&p_cxc->COLMAP.p_quad[i].prim[j]);
			}
		}
	}

	/* render COLMAP's cuboids */
	for(i = 0; i < p_cxc->COLMAP.ncuboids; i++)
	{
		for(j = 0; j < 6; j++)
		{
			for(k = 0; k < 4; k++)
			{
				/* scale */
				v[k].vx = p_cxc->COLMAP.p_cuboid[i].quad[j].global_coords.v[k].vx * (*p_parent_scale);
				v[k].vy = p_cxc->COLMAP.p_cuboid[i].quad[j].global_coords.v[k].vy * (*p_parent_scale);
				v[k].vz = p_cxc->COLMAP.p_cuboid[i].quad[j].global_coords.v[k].vz * (*p_parent_scale);

				/* rotate in model space */
				if(use_pivot_local)
				{
					VECTOR out;

					RotMatrix(p_parent_rotation_model, &model_rot);

					ApplyMatrix(&model_rot, &v[k], &out);
					setVector(&v[k], out.vx, out.vy, out.vz);
				}
				/* translate */
				v[k].vx += p_parent_global_pos->vx + p_global_pos->vx - p_camera->offset.vx;
				v[k].vy += p_parent_global_pos->vy + p_global_pos->vy - p_camera->offset.vy;
				v[k].vz += p_parent_global_pos->vz + p_global_pos->vz - p_camera->offset.vz;
			}

			otz = RotAverage4(&v[0], &v[1], &v[2], &v[3],
				(long *)&p_cxc->COLMAP.p_cuboid[i].quad[j].prim[0].x0,
				(long *)&p_cxc->COLMAP.p_cuboid[i].quad[j].prim[1].x0,
				(long *)&p_cxc->COLMAP.p_cuboid[i].quad[j].prim[3].x0,
				(long *)&p_cxc->COLMAP.p_cuboid[i].quad[j].prim[2].x0,
				&p, &flg);

			/* connect LINE_F2s */
			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[0].x1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[2].x0;
			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[1].x1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[0].x0;
			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[3].x1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[1].x0;
			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[2].x1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[3].x0;

			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[0].y1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[2].y0;
			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[1].y1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[0].y0;
			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[3].y1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[1].y0;
			p_cxc->COLMAP.p_cuboid[i].quad[j].prim[2].y1 = p_cxc->COLMAP.p_cuboid[i].quad[j].prim[3].y0;

			/* either add to OT or not */
			if( use_ot && (otz > MIN_OTZ) && (otz < otsize) )
			{
				for(k = 0; k < 4; k++)
				{
					AddPrim(p_ot+otz, &p_cxc->COLMAP.p_cuboid[i].quad[j].prim[k]);
				}
			}
			else
			{
				for(k = 0; k < 4; k++)
				{
					DrawPrim(&p_cxc->COLMAP.p_cuboid[i].quad[j].prim[k]);
				}
			}
		}
	}

	/* render AREATRIG */
	for(i = 0; i < p_cxc->AREATRIG.ncuboids; i++)
	{
		for(j = 0; j < 6; j++) /* six quads per cuboid */
		{
			for(k = 0; k < 4; k++) /* four vertices per quad */
			{
				/* scale */
				v[k].vx = p_cxc->AREATRIG.p_cuboid[i].quad[j].global_coords.v[k].vx * (*p_parent_scale);
				v[k].vy = p_cxc->AREATRIG.p_cuboid[i].quad[j].global_coords.v[k].vy * (*p_parent_scale);
				v[k].vz = p_cxc->AREATRIG.p_cuboid[i].quad[j].global_coords.v[k].vz * (*p_parent_scale);

				/* rotate in model space */
				if(use_pivot_local)
				{
					VECTOR out;

					RotMatrix(p_parent_rotation_model, &model_rot);

					ApplyMatrix(&model_rot, &v[k], &out);
					setVector(&v[k], out.vx, out.vy, out.vz);
				}
				/* translate */
				v[k].vx += p_parent_global_pos->vx + p_global_pos->vx - p_camera->offset.vx;
				v[k].vy += p_parent_global_pos->vy + p_global_pos->vy - p_camera->offset.vy;
				v[k].vz += p_parent_global_pos->vz + p_global_pos->vz - p_camera->offset.vz;

			}

			otz = RotAverage4(&v[0], &v[1], &v[2], &v[3],
				(long *)&p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[0].x0,
				(long *)&p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[1].x0,
				(long *)&p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[3].x0,
				(long *)&p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[2].x0,
				&p, &flg);

			/* connect LINE_F2s */
			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[0].x1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[2].x0;
			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[1].x1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[0].x0;
			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[3].x1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[1].x0;
			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[2].x1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[3].x0;

			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[0].y1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[2].y0;
			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[1].y1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[0].y0;
			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[3].y1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[1].y0;
			p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[2].y1 = p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[3].y0;

			/* either add to OT or not */
			if( use_ot && (otz > MIN_OTZ) && (otz < otsize) )
			{
				for(k = 0; k < 4; k++)
				{
					AddPrim(p_ot+otz, &p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[k]);
				}
			}
			else
			{
				for(k = 0; k < 4; k++)
				{
					DrawPrim(&p_cxc->AREATRIG.p_cuboid[i].quad[j].prim[k]);
				}
			}
		}
	}

	PopMatrix();
}

/*
 * bezier3d_add:
 *	Adds a 3D bezier curve to an ordering table.
 */
void
bezier3d_add(p_bezier, p_ent, p_player_pos, p_rottrans, rotm, p_camera, flag, p_ot, otsize)
struct Bezier3D *p_bezier;
struct Entity3D *p_ent; /* parent entity */
SVECTOR *p_player_pos;
MATRIX *p_rottrans;
int rotm; /* model-space rotation order */
struct Camera3D *p_camera;
int flag, *p_ot, otsize;
{
	int			i, j, k;

	long			otz, p, flg, isomote;
	int			ndiv;
	short			size;
	MATRIX			model_rot;

	PushMatrix();

	rotmat[rotm](&p_ent->rotation_model, &model_rot);

	RotMatrix_gte(&p_ent->rotation, p_rottrans);
	TransMatrix(p_rottrans, &p_ent->translation);

	gte_SetRotMatrix(p_rottrans);
	gte_SetTransMatrix(p_rottrans);

	/* path prims */
	for(i = 0; i < MAX_BEZIER_PRIM; i++)
	{
		vert_srt(&p_bezier->tilepos[i], NULL, 1, &p_ent->global_pos,
				p_player_pos, &p_camera->offset, 0, /* FIXME: local pivot can be simplified */
				SRT_T, &coord[i]);

		otz = RotTransPers(coord+i, (long *)&p_bezier->tile[i].x0, &p, &flg);
		size = lerpi(8, 0, otz, 12)>>12;
		setWH(&p_bezier->tile[i], size, size);
		if(otz > MIN_OTZ && otz < otsize)
		{
			DrawPrim(&p_bezier->tile[i]);
		}
	}

	/* cursor */
	vert_srt(&p_bezier->cursor.coords, NULL, 1, &p_ent->global_pos,
			p_player_pos, &p_camera->offset, 0, /* FIXME: local pivot can be simplified */
			SRT_T, &coord[0]);

	otz = RotTransPers(coord, (long *)&p_bezier->cursor.tile.x0, &p, &flg);
	size = lerpi(8, 0, otz, 12)>>12;
	setWH(&p_bezier->cursor.tile, size, size);
	if(otz > MIN_OTZ && otz < otsize)
	{
		DrawPrim(&p_bezier->cursor.tile);
	}

	/* keys */
	for(i = 0; i < p_bezier->nkeys; i++)
	{
		vert_srt(&p_bezier->key[i], NULL, 1, &p_ent->global_pos,
				p_player_pos, &p_camera->offset, 0, /* FIXME: local pivot can be simplified */
				SRT_S|SRT_T, &coord[i]);

		otz = RotTransPers(coord+i, (long *)&p_bezier->keyprim[i].x0, &p, &flg);
		size = lerpi(8, 0, otz, 12)>>12;
		setWH(&p_bezier->keyprim[i], size, size);
		if(otz > MIN_OTZ && otz < otsize)
		{
			DrawPrim(&p_bezier->keyprim[i]);
		}
	}

	PopMatrix();
}

/*
 * polylight_init:
 *	Initializes the representation of a point light.
 */
polylight_init(r, g, b)
unsigned char r, g, b;
{
	int i;

	for(i = 0; i < NUMLIGHTS; i++)
	{
		setPolyF4(polylight+i);
		setRGB0(polylight+i, r, g, b);
	}
}

/*
 * dgx_light_point_update:
 *	Applies world translation to active point lights.
 */
dgx_light_point_update(p_player_pos, p_camera)
SVECTOR *p_player_pos;
struct Camera3D *p_camera;
{
	int i;
	SVECTOR zero = {0, 0, 0};

	for(i = 0; i < NUMLIGHTS; i++)
	{
		if(dgx_light_flag & (1<<(i+NUMLIGHTS)))
		{
			vert_srt(dgx_light_point_pos+i, NULL, 1, &zero,
					p_player_pos, &p_camera->offset, 0,
					SRT_S|SRT_T, pointlight+i);
		}
	}
}

/*
 * lod_dist:
 *	Determine whether to render the hi-poly or lo-poly
 *	version of a given model. p_model should be a pointer
 *	to the model's global position.
 */
lod_dist(p_player, p_model)
SVECTOR *p_player;
SVECTOR *p_model;
{
	const MAX_DISTANCE = 45<<FIXED_PRECISION;

	/* if nowhere near model */
	if( (ABS((-p_player->vx) - p_model->vx) > MAX_DISTANCE) ||
		(ABS((-p_player->vy) - p_model->vy) > MAX_DISTANCE) ||
		(ABS((-p_player->vz) - p_model->vz) > MAX_DISTANCE) )
	{
		return LOD_LO;
	}

	return LOD_HI;
}

#include <engine/3d/shapes.h>

/*
 * circle3d_add:
 *	Performs transformation of 3D coords into
 *	screen coords for circles.
 */
circle3d_add(p_ot, p_circle, p_global_pos, p_rottrans, p_camera, use_pivot_local)
unsigned long *p_ot;
struct Circle3D *p_circle;
SVECTOR *p_global_pos;
MATRIX *p_rottrans;
struct Camera3D *p_camera;
int use_pivot_local;
{
	int		i, j;

	MATRIX		model_rot;
	SVECTOR		dmy[3];
	SVECTOR		prev;
	long		p, otz, flg;
	const int	unit = 4096/CIRCLE_LINEF2;

	PushMatrix();

	RotMatrix_gte(&p_circle->ent.rotation, p_rottrans);
	TransMatrix(p_rottrans, &p_circle->ent.translation);

	for(i = 0; i < CIRCLE_LINEF2; i+=3)
	{
		/* rotate in model space */
		if(use_pivot_local)
		{
			RotMatrix_gte(&p_circle->ent.rotation_model, &model_rot);
			for(j = 0; j < 3; j++)
			{
				ApplyMatrixSV(&model_rot, &p_circle->vertex[i+j], &dmy[j]);
			}
		}

		/* translate */
		for(j = 0; j < 3; j++)
		{
			if(!use_pivot_local)
			{
				dmy[j] = p_circle->vertex[i+j];
			}

			dmy[j].vx += (p_circle->ent.global_pos.vx + p_global_pos->vx) - p_camera->offset.vx;
			dmy[j].vy += (p_circle->ent.global_pos.vy + p_global_pos->vy) - p_camera->offset.vy;
			dmy[j].vz += (p_circle->ent.global_pos.vz + p_global_pos->vz) - p_camera->offset.vz;
		}

		gte_SetRotMatrix(p_rottrans);
		gte_SetTransMatrix(p_rottrans);

		otz = RotAverage3(&dmy[0], &dmy[1], &dmy[2],
			(long *)(&p_circle->lines[i].x0), (long *)(&p_circle->lines[i+1].x0),
			(long *)(&p_circle->lines[i+1].x1), &p, &flg);

		p_circle->lines[i].x1 = p_circle->lines[i+1].x0;
		p_circle->lines[i].y1 = p_circle->lines[i+1].y0;

		/* determine visual style */
		switch(p_circle->visual)
		{
			case SOLID:
				if( i == CIRCLE_LINEF2-3 )
				{
					p_circle->lines[i+2].x0 = p_circle->lines[i+1].x1;
					p_circle->lines[i+2].y0 = p_circle->lines[i+1].y1;
					p_circle->lines[i+2].x1 = p_circle->lines[0].x0;
					p_circle->lines[i+2].y1 = p_circle->lines[0].y0;
				}
			case DISCONNECTED:
				if( i != 0 ) 
				{
					p_circle->lines[i-1].x0 = prev.vx;
					p_circle->lines[i-1].y0 = prev.vy;
					p_circle->lines[i-1].x1 = p_circle->lines[i].x0;
					p_circle->lines[i-1].y1 = p_circle->lines[i].y0;
				}

				prev.vx = p_circle->lines[i+1].x1;
				prev.vy = p_circle->lines[i+1].y1;

				break;
		}

		/* add to OT */
		if(otz > MIN_OTZ && otz < game_otsize)
		{
			AddPrim(p_ot+otz, &p_circle->lines[i]);
			AddPrim(p_ot+otz, &p_circle->lines[i+1]);

			switch(p_circle->visual)
			{
				case SOLID:
					/* fit last piece */
					if( i == CIRCLE_LINEF2-3 )
					{
						AddPrim(p_ot+otz, &p_circle->lines[i+2]);
					}
				case DISCONNECTED:
					if( i != 0 )
					{
						AddPrim(p_ot+otz, &p_circle->lines[i-1]);
					}

					break;
			}
		}
	}

	PopMatrix();
}

/*
 * cylinder_add:
 *	Adds a cylinder to an ordering table.
 */
cylinder_add(p_ot, p_gfx, p_camera, p_player_pos, p_rottrans)
unsigned long *p_ot;
struct CylinderGFX *p_gfx;
struct Camera3D *p_camera;
SVECTOR *p_player_pos;
MATRIX *p_rottrans;
{
	int i, j;
	long p, flg;
	SVECTOR zero = {0, 0, 0};
	VECTOR dmy;

	PushMatrix();

	RotMatrix_gte(&p_gfx->ent.rotation, p_rottrans);
	TransMatrix(p_rottrans, &p_gfx->ent.translation);

	gte_SetRotMatrix(p_rottrans);
	gte_SetTransMatrix(p_rottrans);

	for(i = 0; i < 2; i++)
	{
		setVector(&dmy, p_gfx->cyl.p[i].vx, p_gfx->cyl.p[i].vy, p_gfx->cyl.p[i].vz); /* convert SVECTOR to VECTOR */

		vert_srt(&dmy, NULL, 1, &zero,
				p_player_pos, &p_camera->offset, 0,
				SRT_T, &coord[i]);

		RotTransPers(coord+i, (long *)&p_gfx->endp[i].x0, &p, &flg);
		DrawPrim(p_gfx->endp+i);

		p_gfx->endc[i].ent.global_pos = p_gfx->cyl.p[i];
		circle3d_add(p_ot, p_gfx->endc+i, p_player_pos, p_rottrans, p_camera, 0);
	}

	/* glitchy rendering, but I'll leave it since it's just for debugging */
	for(i = 0; i < CYLINDERGFX_BODY_NLINES; i++)
	{
		int idx = i*(CIRCLE_LINEF2/CYLINDERGFX_BODY_NLINES);
		int incr_x = (-p_gfx->endc[0].lines[idx].x0 + p_gfx->endc[1].lines[idx].x0)/CYLINDERGFX_BODY_NSEGMENTS;
		int incr_y = (-p_gfx->endc[0].lines[idx].y0 + p_gfx->endc[1].lines[idx].y0)/CYLINDERGFX_BODY_NSEGMENTS;

		for(j = 0; j < CYLINDERGFX_BODY_NSEGMENTS; j++)
		{
			p_gfx->body[i][j].x0 = p_gfx->endc[0].lines[idx].x0+(j*incr_x);
			p_gfx->body[i][j].y0 = p_gfx->endc[0].lines[idx].y0+(j*incr_y);
			p_gfx->body[i][j].x1 = p_gfx->endc[0].lines[idx].x0+((j+1)*incr_x);
			p_gfx->body[i][j].y1 = p_gfx->endc[0].lines[idx].y0+((j+1)*incr_y);
			DrawPrim(&p_gfx->body[i][j]);
		}
	}
	PopMatrix();
}

unsigned long wf_idx;
LINE_F2 wf_white[WF_WHITE_BUF_SIZE];

/*
 * wireframe_add:
 *	Constructs and renders a wireframe of a given MTD/ATD model.
 *	For this to work properly, make sure you first render the
 *	model itself using the XADD_SAVE_OTZ flag.
 */
wireframe_add(p_asset, type, wf_buf)
void *p_asset;
int type;
LINE_F2 *wf_buf;
{
	int			i, k;
	int			f3idx = 0, f4idx = 0, ft3idx = 0, ft4idx = 0;

	struct APBuf		*p_apb;
	unsigned short		npoly;

	switch(type)
	{
		case ASSET_TYPE_MTD:
		{
			struct AssetMTD *p = (struct AssetMTD *)p_asset;
			p_apb = &p->apb;
			npoly = p->npoly;
			break;
		}
		case ASSET_TYPE_ATD:
		{
			struct AssetATD *p = (struct AssetATD *)p_asset;
			p_apb = &p->apb;
			npoly = p->npoly;
			break;
		}
		default:
			printf("wireframe_add() error: invalid asset type passed: (%d)\n", type);
			while(1);
	}

	/* FIXME: surely there's a better way to write this.. */
	for(i = 0; i < npoly; i++)
	{
		switch(saveotz_type[i])
		{
			case IDX_F3:
				if(saveotz_is_visble[i])
				{
					setXY2(&wf_buf[wf_idx], p_apb->p_polyf3[f3idx].x0, p_apb->p_polyf3[f3idx].y0,
						p_apb->p_polyf3[f3idx].x1, p_apb->p_polyf3[f3idx].y1);

					setXY2(&wf_buf[wf_idx+1], p_apb->p_polyf3[f3idx].x1, p_apb->p_polyf3[f3idx].y1,
						p_apb->p_polyf3[f3idx].x2, p_apb->p_polyf3[f3idx].y2);

					setXY2(&wf_buf[wf_idx+2], p_apb->p_polyf3[f3idx].x2, p_apb->p_polyf3[f3idx].y2,
						p_apb->p_polyf3[f3idx].x0, p_apb->p_polyf3[f3idx].y0);

					for(k = 0; k < 3; k++)
					{
						AddPrim(game_ot+saveotz[i]-1, &wf_buf[wf_idx+k]);
					}
					wf_idx+=3;
				}
				f3idx++;
				break;
			case IDX_F4:
				if(saveotz_is_visble[i])
				{
					setXY2(&wf_buf[wf_idx], p_apb->p_polyf4[f4idx].x0, p_apb->p_polyf4[f4idx].y0,
						p_apb->p_polyf4[f4idx].x1, p_apb->p_polyf4[f4idx].y1);

					setXY2(&wf_buf[wf_idx+1], p_apb->p_polyf4[f4idx].x1, p_apb->p_polyf4[f4idx].y1,
						p_apb->p_polyf4[f4idx].x3, p_apb->p_polyf4[f4idx].y3);

					setXY2(&wf_buf[wf_idx+2], p_apb->p_polyf4[f4idx].x3, p_apb->p_polyf4[f4idx].y3,
						p_apb->p_polyf4[f4idx].x2, p_apb->p_polyf4[f4idx].y2);

					setXY2(&wf_buf[wf_idx+3], p_apb->p_polyf4[f4idx].x2, p_apb->p_polyf4[f4idx].y2,
						p_apb->p_polyf4[f4idx].x0, p_apb->p_polyf4[f4idx].y0);

					for(k = 0; k < 4; k++)
					{
						AddPrim(game_ot+saveotz[i]-1, &wf_buf[wf_idx+k]);
					}
					wf_idx+=4;
				}
				f4idx++;
				break;
			case IDX_FT3:
				if(saveotz_is_visble[i])
				{
					setXY2(&wf_buf[wf_idx], p_apb->p_polyft3[ft3idx].x0, p_apb->p_polyft3[ft3idx].y0,
						p_apb->p_polyft3[ft3idx].x1, p_apb->p_polyft3[ft3idx].y1);

					setXY2(&wf_buf[wf_idx+1], p_apb->p_polyft3[ft3idx].x1, p_apb->p_polyft3[ft3idx].y1,
						p_apb->p_polyft3[ft3idx].x2, p_apb->p_polyft3[ft3idx].y2);

					setXY2(&wf_buf[wf_idx+2], p_apb->p_polyft3[ft3idx].x2, p_apb->p_polyft3[ft3idx].y2,
						p_apb->p_polyft3[ft3idx].x0, p_apb->p_polyft3[ft3idx].y0);

					for(k = 0; k < 3; k++)
					{
						AddPrim(game_ot+saveotz[i]-1, &wf_buf[wf_idx+k]);
					}
					wf_idx+=3;
				}
				ft3idx++;
				break;
			case IDX_FT4:
				if(saveotz_is_visble[i])
				{
					setXY2(&wf_buf[wf_idx], p_apb->p_polyft4[ft4idx].x0, p_apb->p_polyft4[ft4idx].y0,
						p_apb->p_polyft4[ft4idx].x1, p_apb->p_polyft4[ft4idx].y1);

					setXY2(&wf_buf[wf_idx+1], p_apb->p_polyft4[ft4idx].x1, p_apb->p_polyft4[ft4idx].y1,
						p_apb->p_polyft4[ft4idx].x3, p_apb->p_polyft4[ft4idx].y3);

					setXY2(&wf_buf[wf_idx+2], p_apb->p_polyft4[ft4idx].x3, p_apb->p_polyft4[ft4idx].y3,
						p_apb->p_polyft4[ft4idx].x2, p_apb->p_polyft4[ft4idx].y2);

					setXY2(&wf_buf[wf_idx+3], p_apb->p_polyft4[ft4idx].x2, p_apb->p_polyft4[ft4idx].y2,
						p_apb->p_polyft4[ft4idx].x0, p_apb->p_polyft4[ft4idx].y0);

					for(k = 0; k < 4; k++)
					{
						AddPrim(game_ot+saveotz[i]-1, &wf_buf[wf_idx+k]);
					}
					wf_idx+=4;
				}
				ft4idx++;
				break;
		}
	}
}

